from __future__ import annotations
import numpy as np
from pint import UnitRegistry
ureg = UnitRegistry(on_redefinition="ignore")
ureg.enable_contexts("chemistry")
Q_ = ureg.Quantity
ureg.define("Sverdrup = 1e6 * meter **3 / second = Sv = Sverdrups")
ureg.define("Mol = 1 * mol / liter = M")
ureg.define("fraction = [] = frac")
ureg.define("percent = 1e-2 frac = pct")
ureg.define("permil = 1e-3 fraction")
ureg.define("ppm = 1e-6 fraction")
from .esbmtk_base import *
from .esbmtk import *
from .extended_classes import *
from .connections import *
from .processes import *
from .carbonate_chemistry import *
from .utility_functions import *
from .sealevel import *
from .solver import *
from .ode_backend import *
from .ode_backend_2 import *
from .seawater import *
from .carbonate_chemistry import *
from .post_processing import *

np.seterr(invalid="ignore")



"""
esbmtk: A general purpose Earth Science box model toolkit Copyright
(C), 2020 Ulrich G.  Wortmann

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import typing as tp
from esbmtk import Q_, register_return_values
from esbmtk.utility_functions import __addmissingdefaults__, __checkkeys__, __checktypes__

if tp.TYPE_CHECKING:
    from esbmtk import Flux, Model, ReservoirGroup


def add_photosynthesis(
    rgs: list[ReservoirGroup],
    p_fluxes: list,
    piston_velocity,
    O2_At,
    CO2_At,
    CaCO3_reactions=True,
):
    """Add process to ReservoirGroup(s) in rgs. pfluxes must be list of Flux
    objects or float values that correspond to the rgs list
    """
    from esbmtk import register_return_values
    from esbmtk.reactions.init_functions import init_photosynthesis

    M = rgs[0].mo
    pv = piston_velocity.to("meter/yr").magnitude
    for i, rg in enumerate(rgs):
        if isinstance(p_fluxes[i], Q_):
            p_fluxes[i] = p_fluxes[i].to("mol/year").magnitude

        ec = init_photosynthesis(rg, p_fluxes[i], pv, O2_At, CO2_At, CaCO3_reactions)
        register_return_values(ec, rg)
        rg.has_cs1 = True


def add_OM_remineralization(M: Model, f_map: dict) -> None:
    """
    Add OM_remineralization fluxes to the model.

    Parameters:
    M (Model): The model object t
    f_map (dict): A dictionary that maps sink names to source dictionaries. The
    source dictionary should contain the source species and a list of type
    and OM_remineralization values. For example, {M.A_ib: {M.H_sb: ["POM", 0.3]}}.

    Raises:
    ValueError: If an invalid type is specified in the source dictionary.

    Returns:
    None
    """
    from esbmtk.reactions.init_functions import init_OM_remineralization

    # get sink name (e.g., M.A_ib) and source dict e.g. {M.H_sb: {"POM": 0.3}}
    for sink, source_dict in f_map.items():
        pom_fluxes = list()
        pom_fluxes_l = list()
        pom_remin = list()
        pic_fluxes = list()
        pic_fluxes_l = list()
        pic_remin = list()

        # create flux lists for POM and possibly CaCO3
        for source, type_dict in source_dict.items():
            # get matching fluxes for e.g., M.A_sb, and POM
            if "POM" in type_dict:
                fl = M.flux_summary(
                    filter_by=f"photosynthesis {source.name} POM",
                    return_list=True,
                )
                for f in fl:
                    if f.name[-3:] == "F_l":
                        pom_fluxes_l.append(f)
                    else:
                        pom_fluxes.append(f)
                        pom_remin.append(type_dict["POM"])

            if "PIC" in type_dict:
                fl = M.flux_summary(
                    filter_by=f"photosynthesis {source.name} PIC",
                    return_list=True,
                )
                for f in fl:
                    if f.name[-3:] == "F_l":
                        pic_fluxes_l.append(f)
                    else:
                        pic_fluxes.append(f)
                        pic_remin.append(type_dict["PIC"])

        if len(pic_fluxes) > 0:
            ec = init_OM_remineralization(
                sink,
                pom_fluxes,
                pom_fluxes_l,
                pic_fluxes,
                pic_fluxes_l,
                pom_remin,
                pic_remin,
                True,
            )
        else:
            ec = init_OM_remineralization(
                sink,
                pom_fluxes,
                pom_fluxes_l,
                pom_fluxes,  # numba cannot deal with empty fluxes, so we
                pom_fluxes_l,  # add the om_fluxes, but ignore them
                pom_remin,
                pom_remin,
                False,
            )
        register_return_values(ec, sink)
        sink.has_cs2 = True


def add_carbonate_system_3(**kwargs) -> None:
    """Creates a new carbonate system virtual reservoir
    which will compute carbon species, saturation, compensation,
    and snowline depth, and compute the associated carbonate burial fluxes

    Required keywords:
        r_sb: list of ReservoirGroup objects in the surface layer
        r_db: list of ReservoirGroup objects in the deep layer
        pic_export_flux: list of flux objects which must match the
                                 list of ReservoirGroup objects.
        zsat_min = depth of the upper boundary of the deep box
        z0 = upper depth limit for carbonate burial calculations
             typically zsat_min

    Optional Parameters:

        zsat = initial saturation depth (m)
        zcc = initial carbon compensation depth (m)
        zsnow = initial snowline depth (m)
        zsat0 = characteristic depth (m)
        Ksp0 = solubility product of calcite at air-water interface (mol^2/kg^2)
        kc = heterogeneous rate constant/mass transfer coefficient for calcite dissolution (kg m^-2 yr^-1)
        Ca2 = calcium ion concentration (mol/kg)
        pc = characteristic pressure (atm)
        pg = seawater density multiplied by gravity due to acceleration (atm/m)
        I = dissolvable CaCO3 inventory
        co3 = CO3 concentration (mol/kg)
        Ksp = solubility product of calcite at in situ sea water conditions (mol^2/kg^2)

    """
    from esbmtk.reactions.init_functions import init_carbonate_system_3

    # list of known keywords
    lkk: dict = {
        "r_db": list,  # list of deep reservoirs
        "r_ib": list,  # list of corresponding surface reservoirs
        "r_sb": list,  # list of corresponding surface reservoirs
        "pic_export_flux": list,
        "AD": float,
        "zsat": int,
        "zsat_min": int,
        "zcc": int,
        "zsnow": int,
        "zsat0": int,
        "Ksp0": float,
        "kc": float,
        "Ca2": float,
        "pc": (float, int),
        "pg": (float, int),
        "I_caco3": (float, int),
        "alpha": float,
        "zmax": (float, int),
        "z0": (float, int),
        "Ksp": (float, int),
        # "BM": (float, int),
    }
    # provide a list of absolutely required keywords
    lrk: list[str] = [
        "r_db",
        "r_ib",
        "r_sb",
        "pic_export_flux",
        "zsat_min",
        "z0",
    ]

    # we need the reference to the Model in order to set some
    # default values.

    reservoir = kwargs["r_db"][0]
    model = reservoir.mo
    # list of default values if none provided
    lod: dict = {
        "r_sb": [],  # empty list
        "zsat": -3715,  # m
        "zcc": -4750,  # m
        "zsnow": -5000,  # m
        "zsat0": -5078,  # m
        "Ksp0": reservoir.swc.Ksp0,  # mol^2/kg^2
        "kc": 8.84 * 1000,  # m/yr converted to kg/(m^2 yr)
        "AD": model.hyp.area_dz(-200, -6000),
        "alpha": 0.6,  # 0.928771302395292, #0.75,
        "pg": 0.103,  # pressure in atm/m
        "pc": 511,  # characteristic pressure after Boudreau 2010
        "I_caco3": 529,  # dissolveable CaCO3 in mol/m^2
        "zmax": -6000,  # max model depth
        "Ksp": reservoir.swc.Ksp,  # mol^2/kg^2
    }

    # make sure all mandatory keywords are present
    __checkkeys__(lrk, lkk, kwargs)
    # add default values for keys which were not specified
    kwargs = __addmissingdefaults__(lod, kwargs)
    # test that all keyword values are of the correct type
    __checktypes__(lkk, kwargs)

    i = 0
    j = 0
    for rg in kwargs["r_db"]:  # Setup the virtual reservoirs
        ec = init_carbonate_system_3(
            rg,
            kwargs["pic_export_flux"][j],
            kwargs["r_sb"][i],
            kwargs["r_ib"][i],
            kwargs["r_db"][i],
            kwargs,
        )
        i += 1
        j += 2

        register_return_values(ec, rg)
        rg.has_cs2 = True


def add_co2_gas_exchange(rg_list, gas_r):
    """Add gas_exchange. Note that this will register
    the resulting flux only with the GasReservoir. So
    we need to add it to the respective surface reservoir
    manually. Approach. Scan e.c. return objects for the correct flux
    not sure yet how to deal with sign

    :param rg_list: list of reservoir group names
    """
    from esbmtk.reactions.init_functions import init_gas_exchange_with_isotopes

    species = rg_list[0].mo.CO2
    pv = Q_("4.8 m/d")

    for rg in rg_list:
        lr = getattr(rg, "DIC")
        l_ref = getattr(rg, "CO2aq")
        solubility = getattr(lr.swc, "SA_co2")
        ec = init_gas_exchange_with_isotopes(gas_r, lr, l_ref, species, pv, solubility)
        register_return_values(ec, lr)
        rg.has_cs1 = True


def add_o2_gas_exchange(rg_list, gas_r):
    """Add gas_exchange

    :param rg_list: list of reservoir group names
    """
    from esbmtk.reactions.init_functions import init_gas_exchange_no_isotopes

    species = rg_list[0].mo.O2
    pv = Q_("4.8 m/d")

    for rg in rg_list:
        lr = getattr(rg, "O2")
        l_ref = getattr(rg, "O2")
        solubility = getattr(lr.swc, "SA_o2")
        ec = init_gas_exchange_no_isotopes(gas_r, lr, l_ref, species, pv, solubility)
        register_return_values(ec, lr)
        gas_r.has_cs1 = True



"""

     esbmtk.connections

     Classes which handle the connections and fluxes between esbmtk objects
     like Reservoirs, Sources, and Sinks.

     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# from pint import UnitRegistry
from __future__ import annotations
import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
import pandas as pd
import scipy as sp
import scipy.interpolate
from .esbmtk_base import esbmtkBase

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


class hypsometry(esbmtkBase):
    """A class to provide hypsometric data for the depth interval between -6000 to 1000
    meter (relative to sealevel)
    The data is derived from etopo 2, but internally represented by a spline approximation

    Invoke as:
               hyspometry(name="hyp")

    User facing methods:

           hyp.area (z) return the ocean area at a given depth in m^2

           hyp.area_dz(0,-200)

                 will return the surface area between 0 and -200 mbsl (i.e.,
                 the contintal shelves) in percent. This number has a small
                 error since we exclude the areas below 6000 mbsl. The error
                 is however a constant an likely within the uncertainty of
                 the total surface area. The numbers obtained from
                 http://www.physicalgeography.net/fundamentals/8o.html for
                 total ocean area vary between 70.8% for all water covered
                 areas and 72.5% for ocean surface. This routine returns
                 70.5%

           hyp.sa = Earth total surface area in m^2

           hyp.volume(0,200)

           hyp.get_lookup_table(self, min_depth: int, max_depth: int)

                   Generate a vector which contains the area(z) in 1 meter intervals
                   Note that the numbers are area_percentage. To get actual area, you need to_csv
                   multiply with the total surface area (hyp.sa)

    """

    def __init__(self, **kwargs):
        """Initialize a hypsometry object"""

        from esbmtk import Model

        # allowed keywords
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "register": ["None", (Model, str)],
            "model": ["None", (str, Model)],
        }

        # required keywords
        self.lrk: list = [
            "name",
        ]

        self.__initialize_keyword_variables__(kwargs)

        if self.register == "None" and self.model != "None":
            self.register = self.model

        self.parent = self.register

        # legacy variables
        self.pfn = "spline_paramaters.txt"
        self.hfn = "Hypsometric_Curve_05m.csv"
        self.sa = 510067420e6  # total surfce area in square meters, http://www.physicalgeography.net/fundamentals/8o.html
        self.mo = self.model
        self.__register_name_new__()
        self.__init_curve__()
        self.oa = self.area_dz(0, -6000)  # total ocean area

    def volume(self, u: float, l: float) -> float:
        """Calculate the area between two elevation datums

        u = upper limit (e.g., -10)
        l = lower limit (e.g., -100)

        returns the volume in cubic meters
        """

        u = abs(u)
        l = abs(l)
        if l < u:
            raise ValueError(f"hyp.volume: {l} must be higher than {u}")

        return np.sum(self.hypdata[u:l]) * self.sa

    def area(self, depth: int) -> float:
        """Calculate the ocean area at a given depth

        depth must be an integer between 0 and 6000 mbsl, or a
        numpy array of integers between 0 and 6000 mbsl

        """

        depth = np.abs(depth).astype(int)

        if np.max(depth) > 6001:
            raise ValueError("area() is only defined to a depth of 6001 mbsl")

        return np.take(self.hypdata, depth) * self.sa

    def area_dz(self, u: float, l: float) -> float:
        """Calculate the area between two elevation datums

        u = upper limit
        l = lower limit

        the interpolation function returns a numpy array with
        cumulative area percentages do the difference between the
        lowest and highest value is the area contained between
        both limits. The difference between the upper and lower
        bounds is the area percentage contained between both depths.

        The function returns this value multiplied by total surface area,
        i.e., in square meters.

        """

        if l < -6002:
            raise ValueError("area_dz() is only defined to a depth of 6000 mbsl")

        a: NDArrayFloat = sp.interpolate.splev([u, l], self.tck)

        return (a[0] - a[-1]) * self.sa

    def __init_curve__(self):
        """Initialize Spline Parameters. See  __bootstrap_curve__ if you want
        to change the default parameters

        """
        t = [
            -6000.0,
            -6000.0,
            -6000.0,
            -6000.0,
            -5250.0,
            -4500.0,
            -3750.0,
            -3000.0,
            -1500.0,
            -1120.0,
            -750.0,
            -560.0,
            -370.0,
            -180.0,
            -90.0,
            0.0,
            380.0,
            750.0,
            1500.0,
            2250.0,
            3000.0,
            5990.0,
            5990.0,
            5990.0,
            5990.0,
        ]
        c = [
            0.01018464,
            0.00825062,
            0.08976178,
            0.26433525,
            0.44127754,
            0.5799517,
            0.59791548,
            0.6263245,
            0.63035567,
            0.63978284,
            0.64800198,
            0.6501602,
            0.68030866,
            0.75133294,
            0.86590303,
            0.92052208,
            0.96111183,
            0.97330001,
            0.99966578,
            0.99759724,
            1.00067306,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
        k = 3

        self.tck = (t, c, k)

        self.hypdata = sp.interpolate.splev(np.arange(0, -6001, -1), self.tck)

    def __bootstrap_curve__(self):
        """Regenerate the spline data based on the hypsometric data in
        Hypsometric_Curve_05m.csv,

        """
        df = pd.read_csv(
            "Hypsometric_Curve_05m.csv",
            float_precision="high",
            nrows=1200,
            skiprows=300,
        )
        area = df.iloc[:, 2].to_numpy()  # get area as numpy array
        elevation = df.iloc[:, 1].to_numpy()  # get area as numpy array

        tck = sp.interpolate.splrep(
            elevation,
            area,
            s=0.001,
        )
        print(f"t = {tck[0].__repr__()}")
        print(f"c = {tck[1].__repr__()}")
        print(f"k = {tck[2].__repr__()}")

        depth = np.linspace(-6000, 1000, 50)
        a = sp.interpolate.splev(depth, tck)

        plt.style.use(["ggplot"])
        plt.figure()  # Create a figure instance called fig
        ax = plt.subplot()  # Create a plot instance called ax
        ax.plot(elevation, area)  # create a line plot
        ax.plot(depth, a)  # create a line plot
        plt.show()  # display figure

    def get_lookup_table(self, min_depth: int, max_depth: int) -> NDArrayFloat:
        """Generate a vector which contains the area(z) in 1 meter intervals
        The numbers are given in m^2 which represent the actual area.

        The calculations multiply the area_percentage by the total surface area (hyp.sa)
        """

        if not -6002 <= min_depth <= 0:
            raise ValueError("min_depth must be <= 0 and >= -6000")

        if not -6002 <= max_depth <= min_depth:
            raise ValueError("max_depth must be <= 0 and >= -6000")

        return (
            sp.interpolate.splev(np.arange(min_depth, max_depth, -1), self.tck)
            * self.sa
        )

    def get_lookup_table_area_dz(self, min_depth: int, max_depth: int) -> NDArrayFloat:
        """Generate a vector which contains the first derivative of area(z) in 1 meter intervals
        Note that the numbers are in m^2

        """

        return np.diff(self.get_lookup_table(min_depth, max_depth))


def get_box_geometry_parameters(box) -> None:
    from esbmtk import Q_
    import warnings

    box.geometry_unset = True

    if box.geometry != "None":
        if not isinstance(box.geometry, list):
            raise ValueError("geometry must be a list see the docs for details")
        box.area_percentage = box.geometry[2]
        volume = f"{box.mo.hyp.volume(box.geometry[0], box.geometry[1]) * box.area_percentage} m**3"

        box.volume = Q_(volume)
        box.volume = box.volume.to(box.mo.v_unit)
        box.area = box.mo.hyp.area(box.geometry[0]) * box.area_percentage
        box.sed_area = (
            box.mo.hyp.area_dz(box.geometry[0], box.geometry[1]) * box.area_percentage
        )
        box.area_fraction = box.sed_area / box.mo.hyp.oa
        box.area_dz = box.sed_area
        box.geometry_unset = False

        # Define the area_dz property

    elif box.volume == "None":
        raise ValueError("You need to provide volume or geometry!")



"""esbmtk: A general purpose Earth Science box model toolkit Copyright
     (C), 2020-2021 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
     published by the Free Software Foundation, either version 3 of
     the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see
     <https://www.gnu.org/licenses/>.

"""
from __future__ import annotations
import typing as tp
import numpy as np
import numpy.typing as npt
from math import log, sqrt
from .utility_functions import (
    __checkkeys__,
    __addmissingdefaults__,
    __checktypes__,
    register_return_values,
)

if tp.TYPE_CHECKING:
    from .esbmtk import SeawaterConstants, ReservoirGroup

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]

"""
Carbonate System 1 setup requires 3 steps: First we define the actual function,
carbonate_system_1_ode().  In the second step we create a wrapper
init_carbonate_system_1() that defines how to integrate this function into
esbmtk.  In the third step we create a functiom that uses
init_carbonate_system_1() to associates cs1 instances with the respective
reservoirs.

The process for cs2 is analogous
"""


def carbonate_system_1_ode(
    swc: any,
    dic: float,
    dic_l: float,
    ta: float,
    hplus: float,
) -> float:
    """Calculates and returns the H+ and carbonate alkalinity concentrations
     for the given reservoirgroup

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    """

    k1 = swc.K1  # K1
    k2 = swc.K2  # K1
    k1k1 = swc.K1K1  # K1 * K1
    k1k2 = swc.K1K2  # K1 * K2
    KW = swc.KW  # KW
    KB = swc.KB  # KB
    boron = swc.boron  # boron
    hplus_0 = hplus

    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1k1 - k1k2 * (4 - 8 * gamm)
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))
    co2aq: float = dic / (1 + (k1 / hplus) + (k1k2 / (hplus * hplus)))
    co3 = max(dic / (1 + hplus / k2 + hplus * hplus / k1k2), 3.7e-05)
    diff = hplus - hplus_0

    return diff, co2aq, co3


def init_carbonate_system_1(rg: ReservoirGroup):
    """Creates a new carbonate system virtual reservoir for each
    reservoir in rgs. Note that rgs must be a list of reservoir groups.

    Required keywords:
        rgs: list = []  of Reservoir Group objects

    These new virtual reservoirs are registered to their respective Reservoir
    as 'cs'.

    The respective data fields are available as rgs.r.cs.xxx where xxx stands
    for a given key key in the  vr_datafields dictionary (i.e., H, CA, etc.)

    """
    from esbmtk import ExternalCode, carbonate_system_1_ode

    ec = ExternalCode(
        name="cs",
        species=rg.mo.Carbon.CO2,
        function=carbonate_system_1_ode,
        fname="carbonate_system_1_ode",
        ftype="cs1",
        function_input_data=[rg.swc, rg.DIC, rg.TA, "Hplus"],
        register=rg,
        # name and initial value pairs
        # return_values={"Hplus": rg.swc.hplus},
        return_values=[
            {f"R_{rg.full_name}.Hplus": rg.swc.hplus},
            {f"R_{rg.full_name}.CO2aq": rg.swc.co2aq},
            {f"R_{rg.full_name}.CO3": rg.swc.co3},
            # {"Hplus": rg.swc.hplus},
            # {"CO2aq": rg.swc.co2aq},
        ],
        # return_values=["Hplus", "CO2aq"],
    )
    rg.mo.lpc_f.append(ec.fname)

    return ec


def add_carbonate_system_1(rgs: list):
    """Creates a new carbonate system virtual reservoir for each
    reservoir in rgs. Note that rgs must be a list of reservoir groups.

    Required keywords:
        rgs: list = []  of Reservoir Group objects

    These new virtual reservoirs are registered to their respective Reservoir
    as 'cs'.

    The respective data fields are available as rgs.r.cs.xxx where xxx stands
    for a given key key in the  vr_datafields dictionary (i.e., H, CA, etc.)

    """
    from esbmtk import init_carbonate_system_1, Reservoir

    for rg in rgs:
        if hasattr(rg, "DIC") and hasattr(rg, "TA"):
            pass
        else:
            raise AttributeError(f"{rg.full_name} must have a TA and DIC reservoir")

        ec = init_carbonate_system_1(rg)
        register_return_values(ec, rg)
        rg.has_cs1 = True


def carbonate_system_2_ode(
    rg: ReservoirGroup,  # 2 Reservoir handle
    CaCO3_export: float,  # 3 CaCO3 export flux as DIC
    dic_db: float,  # 4 DIC in the deep box
    dic_db_l: float,  # 4 DIC in the deep box
    ta_db: float,  # 5 TA in the deep box
    dic_sb: float,  # 6 [DIC] in the surface box
    dic_sb_l: float,  # 7 [DIC_l] in the surface box
    hplus_0: float,  # 8 hplus in the deep box at t-1
    zsnow: float,  # 9 snowline in meters below sealevel at t-1
    ksp0,
    kc,
    AD,
    zsat0,
    I_caco3,
    alpha,
    zsat_min,
    zmax,
    z0,
) -> tuple:
    """Calculates and returns the fraction of the carbonate rain that is
    dissolved an returned back into the ocean. This functions returns:

    DIC_burial, DIC_burial_l, Hplus, zsnow

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004
    """

    # Parameters
    k1 = rg.swc.K1  # K1
    k1k1 = rg.swc.K1K1
    k2 = rg.swc.K2  # K2
    k1k2 = rg.swc.K1K2  # K2
    KW = rg.swc.KW  # KW
    KB = rg.swc.KB  # KB
    ca2 = rg.swc.ca2  # Ca2+
    boron = rg.swc.boron  # boron
    zsat0 = int(abs(zsat0))
    zsat_min = int(abs(zsat_min))
    zmax = int(abs(zmax))
    z0 = int(abs(z0))
    depth_area_table = rg.cs.depth_area_table
    area_dz_table = rg.cs.area_dz_table
    Csat_table = rg.cs.Csat_table

    # calc carbonate alkalinity based t-1
    oh: float = KW / hplus_0
    boh4: float = boron * KB / (hplus_0 + KB)
    fg: float = hplus_0 - oh - boh4
    ca: float = ta_db + fg

    # calculate carbon speciation
    # The following equations are after Follows et al. 2006
    gamm: float = dic_db / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1k1 - k1k2 * (4 - 8 * gamm)
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))
    co3 = max(dic_db / (1 + hplus / k2 + hplus * hplus / k1k2), 3.7e-05)
    # ---------- compute critical depth intervals eq after  Boudreau (2010)
    # all depths will be positive to facilitate the use of lookup_tables
    zsat = int(zsat0 * log(ca2 * co3 / ksp0))
    zsat = np.clip(zsat, zsat_min, zmax)
    zcc = int(
        zsat0 * log(CaCO3_export * ca2 / (ksp0 * AD * kc) + ca2 * co3 / ksp0)
    )  # eq3
    zcc = np.clip(zcc, zsat_min, zmax)
    # get fractional areas
    B_AD = CaCO3_export / AD
    A_z0_zsat = depth_area_table[z0] - depth_area_table[zsat]
    A_zsat_zcc = depth_area_table[zsat] - depth_area_table[zcc]
    A_zcc_zmax = depth_area_table[zcc] - depth_area_table[zmax]
    # ------------------------Calculate Burial Fluxes----------------------------- #
    BCC = A_zcc_zmax * B_AD
    BNS = alpha * A_z0_zsat * B_AD
    diff_co3 = Csat_table[zsat:zcc] - co3
    area_p = area_dz_table[zsat:zcc]
    BDS_under = kc * area_p.dot(diff_co3)
    BDS_resp = alpha * (A_zsat_zcc * B_AD - BDS_under)
    BDS = BDS_under + BDS_resp
    if zsnow > zmax:
        zsnow = zmax
    diff: NDArrayFloat = Csat_table[zcc : int(zsnow)] - co3
    area_p: NDArrayFloat = area_dz_table[zcc : int(zsnow)]
    # integrate saturation difference over area
    BPDC = kc * area_p.dot(diff)
    BPDC = max(BPDC, 0)  # prevent negative values
    d_zsnow = -BPDC / (area_dz_table[int(zsnow)] * I_caco3)
    BD: float = BDS + BCC + BNS + BPDC

    """CACO3_export is the flux of CaCO3 into the box. However, the model should
    use the bypass option and leave all flux calculations to the
    cs_code.  As such, we simply add the fraction of the input flux
    that dissolves, and ignore the fraction that is buried.  

    The isotope ratio of the dissolution flux is determined by the delta
    value of the sediments we are dissolving, and the delta of the carbonate rain.
    The currrent code, assumes that both are the same.
    """
    BD_l = BD * dic_sb_l / dic_sb
    dH = hplus - hplus_0
    # F_DIC, F_DIC_l, F_TA, dH, d_zsnow
    return BD, BD_l, 2 * BD, dH, d_zsnow, co3


def gas_exchange_ode(scale, gas_c, p_H2O, solubility, g_c_aq) -> float:
    """Calculate the gas exchange flux across the air sea interface

    Parameters:
    scale: surface area in m^2 * piston_velocity
    gas_c: species concentration in atmosphere
    p_H2O: water vapor partial pressure
    solubility: species solubility  mol/(m^3 atm)
    gc_aq: concentration of the dissolved gas in water
    """

    beta = solubility * (1 - p_H2O)
    f = scale * (gas_c * beta - g_c_aq * 1e3)
    return -f


def init_carbonate_system_2(
    rg: ReservoirGroup,
    export_flux: Flux,
    r_sb: ReservoirGroup,
    r_db: ReservoirGroup,
    area_table: NDArrayFloat,
    area_dz_table: NDArrayFloat,
    Csat_table: NDArrayFloat,
    AD: float,
    kwargs: dict,
):
    from esbmtk import ExternalCode, carbonate_system_2_ode

    ec = ExternalCode(
        name="cs",
        species=rg.mo.Carbon.CO2,
        function=carbonate_system_2_ode,
        fname="carbonate_system_2_ode",
        ftype="cs2",
        r_s=r_sb,  # source (RG) of CaCO3 flux,
        r_d=r_db,  # sink (RG) of CaCO3 flux,
        vr_datafields={
            "depth_area_table": area_table,
            "area_dz_table": area_dz_table,
            "Csat_table": Csat_table,
        },
        function_input_data=[
            rg,  # 0
            export_flux,  # 1
            r_db.DIC,  # 2
            r_db.TA,  # 3
            r_sb.DIC,  # 4
            "Hplus",  # 5
            "zsnow",  # 6
            kwargs["Ksp0"],  # 7
            float(kwargs["kc"]),  # 8
            float(AD),  # 9
            float(abs(kwargs["zsat0"])),  # 10
            float(kwargs["I_caco3"]),  # 11
            float(kwargs["alpha"]),  # 12
            float(abs(kwargs["zsat_min"])),  # 13
            float(abs(kwargs["zmax"])),  # 14
            float(abs(kwargs["z0"])),  # 15
        ],
        return_values=[
            {f"F_{rg.full_name}.DIC": "db_cs2"},
            {f"F_{rg.full_name}.TA": "db_cs2"},
            {f"R_{rg.full_name}.Hplus": rg.swc.hplus},
            {f"R_{rg.full_name}.zsnow": float(abs(kwargs["zsnow"]))},
            {f"R_{rg.full_name}.CO3": rg.swc.co3},
        ],
        register=rg,
    )

    return ec


def add_carbonate_system_2(**kwargs) -> None:
    """Creates a new carbonate system virtual reservoir
    which will compute carbon species, saturation, compensation,
    and snowline depth, and compute the associated carbonate burial fluxes

    Required keywords:
        r_sb: list of ReservoirGroup objects in the surface layer
        r_db: list of ReservoirGroup objects in the deep layer
        carbonate_export_fluxes: list of flux objects which must match the
                                 list of ReservoirGroup objects.
        zsat_min = depth of the upper boundary of the deep box
        z0 = upper depth limit for carbonate burial calculations
             typically zsat_min

    Optional Parameters:

        zsat = initial saturation depth (m)
        zcc = initial carbon compensation depth (m)
        zsnow = initial snowline depth (m)
        zsat0 = characteristic depth (m)
        Ksp0 = solubility product of calcite at air-water interface (mol^2/kg^2)
        kc = heterogeneous rate constant/mass transfer coefficient for calcite dissolution (kg m^-2 yr^-1)
        Ca2 = calcium ion concentration (mol/kg)
        pc = characteristic pressure (atm)
        pg = seawater density multiplied by gravity due to acceleration (atm/m)
        I = dissolvable CaCO3 inventory
        co3 = CO3 concentration (mol/kg)
        Ksp = solubility product of calcite at in situ sea water conditions (mol^2/kg^2)

    """

    from esbmtk import Reservoir, init_carbonate_system_2

    # list of known keywords
    lkk: dict = {
        "r_db": list,  # list of deep reservoirs
        "r_sb": list,  # list of corresponding surface reservoirs
        "carbonate_export_fluxes": list,
        "AD": float,
        "zsat": int,
        "zsat_min": int,
        "zcc": int,
        "zsnow": int,
        "zsat0": int,
        "Ksp0": float,
        "kc": float,
        "Ca2": float,
        "pc": (float, int),
        "pg": (float, int),
        "I_caco3": (float, int),
        "alpha": float,
        "zmax": (float, int),
        "z0": (float, int),
        "Ksp": (float, int),
        # "BM": (float, int),
    }
    # provide a list of absolutely required keywords
    lrk: list[str] = [
        "r_db",
        "r_sb",
        "carbonate_export_fluxes",
        "zsat_min",
        "z0",
    ]

    # we need the reference to the Model in order to set some
    # default values.

    reservoir = kwargs["r_db"][0]
    model = reservoir.mo
    # list of default values if none provided
    lod: dict = {
        "r_sb": [],  # empty list
        "zsat": -3715,  # m
        "zcc": -4750,  # m
        "zsnow": -5000,  # m
        "zsat0": -5078,  # m
        "Ksp0": reservoir.swc.Ksp0,  # mol^2/kg^2
        "kc": 8.84 * 1000,  # m/yr converted to kg/(m^2 yr)
        "AD": model.hyp.area_dz(-200, -6000),
        "alpha": 0.6,  # 0.928771302395292, #0.75,
        "pg": 0.103,  # pressure in atm/m
        "pc": 511,  # characteristic pressure after Boudreau 2010
        "I_caco3": 529,  # dissolveable CaCO3 in mol/m^2
        "zmax": -6000,  # max model depth
        "Ksp": reservoir.swc.Ksp,  # mol^2/kg^2
    }

    # make sure all mandatory keywords are present
    __checkkeys__(lrk, lkk, kwargs)

    # add default values for keys which were not specified
    kwargs = __addmissingdefaults__(lod, kwargs)

    # test that all keyword values are of the correct type
    __checktypes__(lkk, kwargs)

    # establish some shared parameters
    # depths_table = np.arange(0, 6001, 1)
    depths: NDArrayFloat = np.arange(0, 6002, 1, dtype=float)
    r_db = kwargs["r_db"]
    r_sb = kwargs["r_sb"]
    ca2 = r_db[0].swc.ca2
    pg = kwargs["pg"]
    pc = kwargs["pc"]
    z0 = kwargs["z0"]
    Ksp0 = kwargs["Ksp0"]
    # test if corresponding surface reservoirs have been defined
    if len(r_sb) == 0:
        raise ValueError(
            "Please update your call to add_carbonate_system_2 and add the list of of corresponding surface reservoirs"
        )

    # C saturation(z) after Boudreau 2010
    Csat_table: NDArrayFloat = (Ksp0 / ca2) * np.exp((depths * pg) / pc)
    area_table = model.hyp.get_lookup_table(0, -6002)  # area in m^2(z)
    area_dz_table = model.hyp.get_lookup_table_area_dz(0, -6002) * -1  # area'
    AD = model.hyp.area_dz(z0, -6000)  # Total Ocean Area

    for i, rg in enumerate(r_db):  # Setup the virtual reservoirs
        ec = init_carbonate_system_2(
            rg,
            kwargs["carbonate_export_fluxes"][i],
            r_sb[i],
            r_db[i],
            area_table,
            area_dz_table,
            Csat_table,
            AD,
            kwargs,
        )

        register_return_values(ec, rg)
        rg.has_cs2 = True


def gas_exchange_ode_with_isotopes(
    scale,  # surface area in m^2 * piston velocity
    gas_c,  # species concentration in atmosphere
    gas_c_l,  # same but for the light isotope
    liquid_c,  # c of the reference species (e.g., DIC)
    liquid_c_l,  # same but for the light isotopeof DIC
    p_H2O,  # water vapor pressure
    solubility,  # solubility constant
    gas_c_aq,  # Gas concentration in liquid phase
    a_db,  # fractionation factor between dissolved CO2aq and HCO3-
    a_dg,  # fractionation between CO2aq and CO2g
    a_u,  # kinetic fractionation during gas exchange
) -> tuple(float, float):
    """Calculate the gas exchange flux across the air sea interface
    for co2 including isotope effects.

    Note that the sink delta is co2aq as returned by the carbonate VR
    this equation is for mmol but esbmtk uses mol, so we need to
    multiply by 1E3

    The Total flux across interface dpends on the difference in either
    concentration or pressure the atmospheric pressure is known, as gas_c, and
    we can calculate the equilibrium pressure that corresponds to the dissolved
    gas in the water as [CO2]aq/beta.

    Conversely, we can convert the the pCO2 into the amount of dissolved CO2 =
    pCO2 * beta

    The h/c ratio in HCO3 estimated via h/c in DIC. Zeebe writes C12/C13 ratio
    but that does not work. the C13/C ratio results however in -8 permil
    offset, which is closer to observations
    """
    # Solibility with correction for pH2O
    beta = solubility * (1 - p_H2O)
    # f as afunction of solubility difference
    f = scale * (beta * gas_c - gas_c_aq * 1e3)
    # isotope ratio of DIC
    Rt = (liquid_c - liquid_c_l) / liquid_c
    # get heavy isotope concentrations in atmosphere
    gas_c_h = gas_c - gas_c_l  # gas heavy isotope concentration
    # get exchange of the heavy isotope
    f_h = scale * a_u * (a_dg * gas_c_h * beta - Rt * a_db * gas_c_aq * 1e3)
    f_l = f - f_h  # the corresponding flux of the light isotope
    # print(f"scale = {scale:2e}")
    # print(f"beta = {beta:2e}")
    # print(f"gas_c = {gas_c:2e}")
    # print(f"gas_c_aq = {gas_c_aq:2e}")
    # print(f"liquid_c = {liquid_c:2e}")
    # print(f"liquid_c_l = {liquid_c_l:2e}")
    # print(f"gas_c_h = {gas_c_h:2e}")
    # print(f"f = {f:2e}")
    # print(f"f_l = {f_l:2e}")
    # breakpoint()

    return -f, -f_l


def get_pco2(SW) -> float:
    """Calculate the concentration of pCO2"""

    dic_c: float = SW.dic
    hplus_c: float = SW.hplus
    k1: float = SW.K1
    k2: float = SW.K2
    co2: NDArrayFloat = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))
    pco2: NDArrayFloat = co2 / SW.K0 * 1e6
    return pco2


def calc_pCO2(
    dic,  # see above why no type hints
    hplus,
    SW,
) -> NDArrayFloat:
    """
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic: Reservoir  = DIC concentrations in mol/kg
    hplus: Reservoir = H+ concentrations in mol/kg
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2(dic,h,SW)

     DataField(name = "SurfaceWaterpCO2",
                       associated_with = reservoir_handle,
                       y1_data = pco2,
                       y1_label = r"pCO_{2}",
                       y1_legend = r"pCO_{2}",
                       )
    Author: T. Tsan

    """

    dic_c: NDArrayFloat = dic.c
    hplus_c: NDArrayFloat = hplus.c

    k1: float = SW.K1
    k2: float = SW.K2

    co2: NDArrayFloat = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))

    pco2: NDArrayFloat = co2 / SW.K0 * 1e6

    return pco2


def calc_pCO2b(
    dic: NDArrayFloat,
    hplus: NDArrayFloat,
    SW: SeawaterConstants,
) -> NDArrayFloat:
    """
    Same as calc_pCO2, but accepts values/arrays rather than Reservoirs.
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic:  = DIC concentrations in mol/kg
    hplus: = H+ concentrations in mol/kg
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2b(dic,h,SW)
     DataField(name = "SurfaceWaterpCO2",
                       associated_with = reservoir_handle,
                       y1_data = pco2b,
                       y1_label = r"pCO_{2}",
                       y1_legend = r"pCO_{2}",
                       )
    """

    dic_c: NDArrayFloat = dic

    hplus_c: NDArrayFloat = hplus

    k1: float = SW.K1
    k2: float = SW.K2

    co2: NDArrayFloat = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))

    pco2: NDArrayFloat = co2 / SW.K0 * 1e6

    return pco2


# define a transform function to display the Hplus concentration as pH
def phc(m: float) -> float:
    """the reservoir class accepts a plot transform. here we use this to
    display the H+ concentrations as pH. After import, you can use it
    with like this in the reservoir definition

     plot_transform_c=phc,

    """
    import numpy as np

    pH = -np.log10(m)
    return pH


def get_hplus(
    rg: ReservoirGroup,
    dic: float,
    ta: float,
    hplus: float,
) -> float:
    k1 = rg.swc.K1  # K1
    k2 = rg.swc.K2  # K2
    KW = rg.swc.KW  # KW
    KB = rg.swc.KB  # KB
    boron = rg.swc.boron  # boron
    hplus_0 = hplus

    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1 * k1 - 4 * k1 * k2 * (1 - (2 * gamm))
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))

    return hplus - hplus_0



from __future__ import annotations
import numpy as np
import numpy.typing as npt
# from numba.typed import List
from esbmtk import Flux, Reservoir, Model, Connection, Connect
from esbmtk import AirSeaExchange

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


def get_initial_conditions(
    M: Model,
    rtol: float,
    atol_d: float = 1e-7,
) -> tuple[list, dict, list, list, NDArrayFloat]:
    """Get list of initial conditions.  This list needs to match the
    number of equations.

    :param Model: The model handle
    :param rtol: relative tolerance for BDF solver.
    :param atol_d: default value for atol if c = 0

    :return: R = list of initial conditions as floats
    :return: icl = dict[Reservoir, list[int, int]] where reservoir
             indicates the reservoir handle, and the list contains the
             index into the reservoir data.  list[0] = concentration
             list[1] concentration of the light isotope.
    :return: cpl = list of reservoirs that use function to evaluate
             reservoir data
    :return: ipl = list of static reservoirs that serve as input
    :return: rtol = array of tolerence values for ode solver

        We need to consider 3 types of reservoirs:

        1) Reservoirs that change as a result of physical fluxes i.e.
        r.lof > 0.  These require a flux statements and a reservoir
        equation.

        2) Reservoirs that do not have active fluxes but are computed
        as a tracer, i.e.. HCO3.  These only require a reservoir
        equation

        3) Reservoirs that do not change but are used as input.  Those
        should not happen in a well formed model, but we cannot
        exclude the possibility.  In this case, there is no flux
        equation, and we state that dR/dt = 0

        get_ic() will look up the index position of the
        reservoir_handle on icl, and then use this index to retrieve
        the correspinding value in R

        Isotopes are handled by adding a second entry
    """
    import numpy as np

    R = []  # list of initial conditions
    atol: list = []  # list of tolerances for ode solver
    # dict that contains the reservoir_handle as key and the index positions
    # for c and l as a list
    icl: dict[Reservoir, list[int, int]] = {}
    cpl: list = []  # list of reservoirs that are computed
    ipl: list = []  # list of static reservoirs that serve as input

    i: int = 0
    for r in M.lic:
        # collect all reservoirs that have initial conditions
        # if r.rtype != "flux_only":
        if len(r.lof) > 0 or r.rtype == "computed" or r.rtype == "passive":
            R.append(r.c[0])  # add initial condition
            if r.c[0] > 0:
                # compute tol such that tol < rtol * abs(y)
                tol = rtol * abs(r.c[0]) / 10
                atol.append(tol)
                r.atol[0] = tol
            else:
                atol.append(atol_d)
                r.atol[0] = atol_d

            if r.isotopes:
                if r.l[0] > 0:
                    # compute tol such that tol < rtol * abs(y)
                    tol = rtol * abs(r.l[0]) / 10
                    atol.append(tol)
                    r.atol[1] = tol
                else:
                    atol.append(atol_d)
                    r.atol[1] = atol_d

                R.append(r.l[0])  # add initial condition for l
                icl[r] = [i, i + 1]
                i += 2
            else:
                icl[r] = [i, i]
                i += 1
            # if r.name == "O2_At":
            #     breakpoint()

    return R, icl, cpl, ipl, np.array(atol)


def write_reservoir_equations(eqs, M: Model, rel: str, ind2: str, ind3: str) -> str:
    """Loop over reservoirs and their fluxes to build the reservoir
    equation

    :param eqs: equation file handle
    :param rel: string with reservoir names used in return function.
        Note that these are the reervoir names as used by the
        equations and not the reservoir names used by esbmtk. E.g.,
        M1.R1.O2 will be M1_R1_O2,
    :param ind2: string with indentation offset
    :param ind3: string with indentation offset

    :returns: rel = updated list of reservoirs names
    """

    for r in M.lor:  # loop over reservoirs
        """
        Write equations for each reservoir and create unique variable
        names.  Reservoirs are typiclally called M.rg.r so we replace
        all dots with underscore -> M_rg_r
        """

        if r.rtype != "flux_only":
            name = f'dCdt_{r.full_name.replace(".", "_")}'
            fex = ""
            v_val = f"{r.volume.to(r.v_unit).magnitude}"

            # add all fluxes
            for flux in r.lof:  # check if in or outflux
                if flux.parent.source == r:
                    sign = "-"
                else:
                    # elif flux.parent.sink == r:
                    sign = "+"

                fname = f'{flux.full_name.replace(".", "_")}'
                fex = f"{fex}{ind3}{sign} {fname}\n"

            if len(r.lof) > 0:  # avoid reservoirs without active fluxes
                if r.ef_results:
                    eqs.write(f"{ind2}{name} += (\n{fex}{ind2})/{v_val}\n\n")
                else:
                    eqs.write(f"{ind2}{name} = (\n{fex}{ind2})/{v_val}\n\n")

                rel = f"{rel}{ind3}{name},\n"

    return rel


def write_reservoir_equations_with_isotopes(
    eqs,
    M: Model,
    rel: str,  # string with reservoir names returned by setup_ode
    ind2: str,  # indent 2 times
    ind3: str,  # indent 3 times
) -> str:  # updated list of reservoirs
    """Loop over reservoirs and their fluxes to build the reservoir equation"""

    for r in M.lor:  # loop over reservoirs
        if r.rtype != "flux_only":
            v_val = f"{r.volume.to(r.v_unit).magnitude}"
            # Write equations for each reservoir
            # create unique variable names. Reservoirs are typiclally called
            # M.rg.r so we replace all dots with underscore
            if r.isotopes:
                name = f'dCdt_{r.full_name.replace(".", "_")}_l'
                fex = ""
                # add all fluxes
                for flux in r.lof:  # check if in or outflux
                    if flux.parent.source == r:
                        sign = "-"
                    else:
                        # elif flux.parent.sink == r:
                        sign = "+"

                    fname = f'{flux.full_name.replace(".", "_")}_l'
                    fex = f"{fex}{ind3}{sign} {fname}\n"

                # avoid reservoirs without active fluxes
                if len(r.lof) > 0:
                    if r.ef_results:
                        eqs.write(f"{ind2}{name} += (\n{fex}{ind2})/{v_val}\n\n")
                    else:
                        eqs.write(f"{ind2}{name} = (\n{fex}{ind2})/{v_val}\n\n")

                    rel = f"{rel}{ind3}{name},\n"

    return rel


def write_equations_2(
    M: Model, R: list[float], icl: dict, cpl: list, ipl: list
) -> tuple:
    """Write file that contains the ode-equations for the Model
    Returns the list R that contains the initial condition for each
    reservoir

    :param Model: Model handle
    :param R: list of floats with the initial conditions for each
              reservoir
    :param icl: dict of reservoirs that have actual fluxes
    :param cpl: list of reservoirs that have no fluxes but are
        computed based on other reservoirs
    :param ipl: list of reservoir that do not change in concentration
    """
    # from esbmtk import Model, ReservoirGroup
    import pathlib as pl

    # get pathlib object
    fn: str = "equations.py"  # file name
    cwd: pl.Path = pl.Path.cwd()  # get the current working directory
    fqfn: pl.Path = pl.Path(f"{cwd}/{fn}")  # fully qualified file name

    # construct header and static code:
    header = """from __future__ import annotations\n\n
from numpy import array as npa
from esbmtk import carbonate_system_1_ode, carbonate_system_2_ode
from esbmtk import gas_exchange_ode, gas_exchange_ode_with_isotopes

class setup_ode():
    '''Class stub to enable state in the equation system passed to ODEINT
    '''
    

    def __init__(self, M: Model)->None:
        ''' Use this method to initialize all variables that require the state
            t-1
        '''

    def eqs(self, t, R: list, M: Model, area_table, area_dz_table, Csat_table) -> list:
        '''Auto generated esbmtk equations do not edit
        '''

"""

    ind2 = 8 * " "  # indention
    ind3 = 12 * " "  # indention
    rel = ""  # list of return values

    # add optional import statements
    if len(M.lpc_f) > 0:
        if M.custom_functions:
            hi = f"{ind2}from custom_functions import "
            for f in set(M.lpc_f):
                hi += f"{f} ,"
        elif M.default_reactions:
            hi = f"{ind2}from esbmtk.reactions.bio_geochemical_reactions import "
            for f in set(M.lpc_f):
                hi += f"{f} ,"
        else:
            hi = f"{ind2}from esbmtk.experimental_reactions.bio_geochemical_reactions import "
            for f in set(M.lpc_f):
                hi += f"{f} ,"
    else:
        hi = "  "

    hi = hi[0:-2]
    header += f"{hi} \n\n"

    # """
    # write file
    with open(fqfn, "w", encoding="utf-8") as eqs:
        eqs.write(header)
        sep = (
            "# ---------------- write computed reservoir equations -------- #\n"
            + "# that do not depend on fluxes"
        )
        eqs.write(f"\n{sep}\n")

        for r in M.lpc_r:  # All virtual reservoirs need to be in this list
            if r.ftype == "cs1":
                # rel = write_cs_1(eqs, r, icl, rel, ind2, ind3)
                rel = write_ef(eqs, r, icl, rel, ind2, ind3)
            elif r.ftype != "cs2":
                raise ValueError(f"{r.ftype} is undefined")

        flist = []
        sep = "# ---------------- write all flux equations ------------------- #"
        eqs.write(f"\n{sep}\n")
        for flux in M.lof:  # loop over fluxes
            if flux.register.ctype == "ignore":
                continue  # skip
            # fluxes belong to at least 2 reservoirs, so we need to avoid duplication
            # we cannot use a set, since we need to preserve order
            if flux not in flist:
                flist.append(flux)  # add to list of fluxes already computed

                if isinstance(flux.parent, Reservoir):
                    continue  # skip computed fluxes

                ex, exl = get_flux(flux, M, R, icl)  # get flux expressions
                fn = flux.full_name.replace(".", "_")

                # check for types that return isotope data as well
                if isinstance(flux.parent, AirSeaExchange):
                    if flux.parent.isotopes:  # add F_l if necessary
                        fn = f"{fn}, {fn}_l"
                    eqs.write(f"{ind2}{fn} = {ex}\n")

                else:  # all others types
                    eqs.write(f"{ind2}{fn} = {ex}\n")
                    if flux.parent.isotopes:  # add line for isotopes
                        eqs.write(f"{ind2}{fn}_l = {exl}\n")

        sep = (
            "# ---------------- write computed reservoir equations -------- #\n"
            + "# that do depend on fluxes"
        )

        eqs.write(f"\n{sep}\n")

        for r in M.lpc_r:  # All virtual reservoirs need to be in this list
            if r.ftype == "cs1":
                pass  # see above
            elif r.ftype == "cs2":  #
                # rel = write_cs_2(eqs, r, icl, rel, ind2, ind3)
                rel = write_ef(eqs, r, icl, rel, ind2, ind3)
            else:
                raise ValueError(f"{r.ftype} is undefined")

        sep = "# ---------------- write input only reservoir equations -------- #"
        eqs.write(f"\n{sep}\n")
        for r in ipl:
            rname = r.full_name.replace(".", "_")
            eqs.write(f"{ind2}{rname} = 0.0")

        sep = "# ---------------- write regular reservoir equations ------------ #"
        eqs.write(f"\n{sep}\n")

        # rel = write_reservoir_equations(eqs, M, rel, ind2, ind3)
        rel = write_reservoir_equations(eqs, M, rel, ind2, ind3)

        sep = "# ---------------- write isotope reservoir equations ------------ #"
        eqs.write(f"\n{sep}\n")

        # rel = write_reservoir_equations(eqs, M, rel, ind2, ind3)
        rel = write_reservoir_equations_with_isotopes(eqs, M, rel, ind2, ind3)

        sep = "# ---------------- bits and pieces --------------------------- #"
        eqs.write(f"\n{sep}\n" f"{ind2}return [\n")
        # Write all initial conditions that are recorded in icl
        for k, v in icl.items():
            eqs.write(f"{ind3}dCdt_{k.full_name.replace('.', '_')},  # {v[0]}\n")
            if k.isotopes:
                eqs.write(f"{ind3}dCdt_{k.full_name.replace('.', '_')}_l,  # {v[1]}\n")

        eqs.write(f"{ind2}]\n")

    return fqfn


def get_flux(flux: Flux, M: Model, R: list[float], icl: dict) -> tuple(str, str):
    """Create formula expressions that calcultes the flux F.  Return
    the equation expression as string

    :param flux: The flux object for which we create the equation
    :param M: The current model object
    :param R: The list of initial conditions for each reservoir
    :param icl: dict of reservoirs that have actual fluxes

    :returns: A tuple where the first string is the equation for the
              total flux, and the second string is the equation for
              the flux of the light isotope
    """
    ind2 = 8 * " "  # indentation
    ind3 = 12 * " "  # indentation

    ex = ""  # expression string
    exl = ""  # expression string for light isotope
    c = flux.parent  # shorthand for the connection object
    cfn = flux.parent.full_name  # shorthand for the connection object name

    if c.ctype.casefold() == "regular":
        ex, exl = get_regular_flux_eq(flux, c, icl, ind2, ind3)

    elif c.ctype == "scale_with_concentration":
        ex, exl = get_scale_with_concentration_eq(flux, c, cfn, icl, ind2, ind3)

    elif c.ctype == "scale_with_mass":
        ex, exl = get_scale_with_concentration_eq(flux, c, cfn, icl, ind2, ind3)

    elif c.ctype == "scale_with_flux":
        ex, exl = get_scale_with_flux_eq(flux, c, cfn, icl, ind2, ind3)

    elif c.ctype == "weathering":
        ex, exl = get_weathering_eq(flux, c, cfn, icl, ind2, ind3)

    elif c.ctype == "gas_exchange":  # Gasexchange
        if c.isotopes:
            ex = get_gas_exchange_w_isotopes_eq(flux, c, cfn, icl, ind2, ind3)
        else:
            ex = get_gas_exchange_eq(flux, c, cfn, icl, ind2, ind3)

    elif c.ctype == "ignore":
        pass  # do nothing

    else:
        raise ValueError(
            f"Connection type {c.ctype} for {c.full_name} is not implmented"
        )

    return ex, exl


def check_signal_2(ex: str, exl: str, c: Connection) -> (str, str):
    """Test if connection is affected by a signal

    :param ex: equation string
    :param c: connection object

    :returns: (modified) equation string
    """

    if c.signal != "None":  # get signal type
        if c.signal.stype == "addition":
            sign = "+"
        elif c.signal.stype == "multiplication":
            sign = "*"
        else:
            raise ValueError(f"stype={c.signal.stype} not implemented")

        ex += f" {sign} {c.signal.full_name}(t)[0]  # Signal"
        if c.source.isotopes:
            exl += f" {sign} {c.signal.full_name}(t)[1]  # Signal"
        else:
            exl += ""

    return ex, exl


def get_ic(r: Reservoir, icl: dict, isotopes=False) -> str:
    """Get initial condition in a reservoir.  If the reservoir is icl,
    return index expression into R.c. If the reservoir is not in the
    index, return the Reservoir concentration a t=0

    In both cases return these a string

    If isotopes == True, return the pointer to the light isotope
    concentration

    :param r: A reservoir handle
    :param icl: icl = dict[Reservoir, list[int, int]] where reservoir
        indicates the reservoir handle, and the list contains the
        index into the reservoir data.  list[0] = concentration
        list[1] concentration of the light isotope.

    :raises ValueError: get_ic: can't find {r.full_name} in list of
        initial conditions

    :returns: the string s which is the full_name of the reservoir
              concentration or isotope concentration
    """
    from esbmtk import Source, Sink

    if r in icl:
        s1 = f"R[{icl[r][0]}]"
        if isotopes:
            s1 += f", R[{icl[r][1]}]"

    elif isinstance(r, (Source, Sink)):
        s1 = f"{r.full_name}.c"
        if isotopes:
            s1 += f", {r.full_name}.l"
    else:
        raise ValueError(
            f"get_ic: can't find {r.full_name} in list of initial conditions"
        )

    return s1


def parse_esbmtk_return_data_types():
    pass


def parse_esbmtk_return_data_types_old(
    line: any, r: Reservoir, ind: str, icl: dict
) -> str:
    """Parse esbmtk data types that are provided as arguments
    to external function objects, and convert them into a suitable string
    format that can be used in the ode equation file
    """
    from esbmtk import GasReservoir, ReservoirGroup
    from .utility_functions import get_reservoir_reference

    # convert to object handle if need be
    if isinstance(line, str):
        o = getattr(r.register, line)
    elif isinstance(line, dict):
        k = next(iter(line))  # get first key
        v = line[k]
        # get reservoir and species handle
        r, sp = get_reservoir_reference(k, v, r.mo)
        if k[:2] == "F_":  # this is a flux
            if isinstance(r, (GasReservoir, Reservoir)):
                sr = f"{r.register.name}.{r.name}.{v}_F".replace(".", "_")
                #  sr = f"M.{r.register.name}_F".replace(".", "_")
                o = r
            elif isinstance(r, ReservoirGroup):
                sr = f"{r.full_name}.{sp.name}.{r.name}.{line[k]}_F".replace(".", "_")
                o = r
            else:
                raise ValueError(f"r = {type(r)}")

        elif k[:2] == "R_":  # this is reservoir or reservoirgroup
            if isinstance(r, (GasReservoir, Reservoir)):
                sr = f"{r.full_name}".replace(".", "_")
                o = r
            elif isinstance(r, ReservoirGroup):
                sr = f"{r.full_name}.{sp.name}".replace(".", "_")
                o = r
            else:
                raise ValueError(f"r = {type(r)}")

    elif isinstance(line, Flux):
        o = line
        sr = o.full_name.replace(".", "_")
    else:  # argument is a regular reservoir
        o = line
        sr = f'dCdt_{o.full_name.replace(".", "_")}'

    if isinstance(o, (Reservoir, GasReservoir, Flux)):
        isotopes = o.isotopes
    else:
        isotopes = getattr(o, sp.name).isotopes
    if isotopes:
        sr = f"{sr}, {sr}_l"

    return sr


def parse_esbmtk_input_data_types(d: any, r: Reservoir, ind: str, icl: dict) -> str:
    """Parse esbmtk data types that are provided as arguments
    to external function objects, and convert them into a suitable string
    format that can be used in the ode equation file
    """
    from esbmtk import Flux, Reservoir, ReservoirGroup, SeawaterConstants, Q_
    from esbmtk import GasReservoir

    if isinstance(d, str):
        sr = getattr(r.register, d)
        a = f"{ind}{get_ic(sr, icl)},\n"
    elif isinstance(d, Reservoir):
        a = f"{ind}{get_ic(d, icl,d.isotopes)},\n"
    elif isinstance(d, GasReservoir):
        a = f"{ind}{get_ic(d, icl,d.isotopes)},\n"
    elif isinstance(d, ReservoirGroup):
        a = f"{ind}{d.full_name},\n"
    elif isinstance(d, Flux):
        sr = d.full_name.replace(".", "_")
        a = f"{ind}{sr},\n"
    elif isinstance(d, SeawaterConstants):
        a = f"{ind}{d.full_name},\n"
    elif isinstance(d, (float, int, np.ndarray)):
        a = f"{ind}{d},\n"
    elif isinstance(d, dict):
        # get key and reservoiur handle
        sr = getattr(r.register, next(iter(d)))
        a = f"{ind}{get_ic(sr, icl)},\n"
    elif isinstance(d, Q_):
        a = f"{ind}{d.magnitude},\n"
    elif isinstance(d, float):
        a = f"{ind}{d.magnitude},\n"
    elif isinstance(d, list):  # loo pover list elements
        a = f"{ind}{d},\n"
        a = f"{ind}npa(["
        for e in d:
            a += f"{parse_esbmtk_input_data_types(e, r,'',icl)[0:-2]},"
        a += "]),\n"
    else:
        raise ValueError(
            f"\n r = {r.full_name}, d ={d}\n" f"\n{d} is of type {type(d)}\n",
        )

    return a


def parse_function_params(params, ind) -> str:
    """Parse function_parameters and convert them into a suitable string
    format that can be used in the ode equation file
    """
    a = ""
    for p in params:
        if isinstance(p, str):
            a += f"{ind}{p},\n"
        else:
            a += f"{ind}{p},\n"
    return a


# ------------------------ define processes ------------------------- #
def write_ef(eqs, r: Reservoir, icl: dict, rel: str, ind2: str, ind3: str) -> str:
    """Write external function call code

    :param eqs: equation file handle
    :param r: reservoir_handle
    :param icl: dict of reservoirs that have actual fluxes
    :param rel: string with reservoir names returned by setup_ode
    :param ind2: indent 2 times
    :param ind3: indent 3 times

    :returns: rel: modied string of reservoir names
    """

    # get list of return values. This is a bit of hack, since we imply
    # anything that its not a flux, is a reservoir with no other inputs
    rv = ind2
    for o in r.lro:
        if isinstance(o, Flux):
            rv += f"{o.full_name.replace('.', '_')}, "
        else:
            rv += f"dCdt_{o.full_name.replace('.', '_')}, "

    rv = rv[:-2]

    a = ""
    # this can probably be simplified simiar to the old parse_return_values()
    for d in r.function_input_data:
        a += parse_esbmtk_input_data_types(d, r, ind3, icl)

    if r.function_params == "None":
        eqs.write(f"{rv} = {r.fname}(\n{a}{ind2})\n\n")
    else:
        params = parse_function_params(r.function_params, ind3)
        eqs.write(f"{rv} = {r.fname}(\n{a}{ind3}(\n{params}{ind3}),\n{ind2})\n\n")
    rel += f"{ind3}{rv},\n"

    return rel


def get_regular_flux_eq(
    flux: Flux,  # flux instance
    c: Connection,  # connection instance
    icl: dict,  # list of initial conditions
    ind2,  # indentation
    ind3,  # indentation
) -> tuple:
    """Create a string containing the equation for a regular (aka
    fixed rate) connection

    :param flux: flux instance
    :param c: connection object
    :param icl: dict of reservoirs that have actual fluxes
    :param ind2: indent 2 times
    :param ind3: indent 3 times

    :returns: two strings, where the first describes the equation for
              the total flux, and the second describes the rate for
              the light isotope
    """
    ex = f"{flux.full_name}.rate"  # get flux rate string
    exl = check_isotope_effects(ex, c, icl, ind3, ind2)
    ex, exl = check_signal_2(ex, exl, c)  # check if we hav to add a signal

    return ex, exl


def check_isotope_effects(
    f_m: str,
    c: Connection,
    icl: dict,
    ind3: str,
    ind2: str,
) -> str:
    """Test if the connection involves any isotope effects

    :param f_m: string with the flux name
    :param c: connection object
    :param icl: dict of reservoirs that have actual fluxes
    :param ind2: indent 2 times
    :param ind3: indent 3 times

    :returns eq: equation string
    """
    if c.isotopes:
        r = c.source.species.r
        s = get_ic(c.source, icl, True)
        s_c, s_l = s.replace(" ", "").split(",")
        """ Calculate the flux of the light isotope (f_l) as a function of the isotope
        ratios in the source reservoir soncentrations (s_c, s_l), and alpha (a) as
        f_l = f_m * 1000/(r * (d + 1000) + 1000)

        Note that the scale has already been applaied to f_m in the calling function.
        """
        if c.delta != "None":
            d = c.delta
            eq = f"{f_m} * 1000 / ({r} * ({d} + 1000) + 1000)"
        elif c.alpha != "None":
            a = c.alpha / 1000 + 1
            eq = f"{s_l} * {f_m} / ({a} * {s_c} + {s_l} - {a} * {s_l})"
        else:
            eq = f"{f_m} * {s_l} / {s_c}"
    else:
        eq = ""

    return eq


def get_scale_with_concentration_eq(
    flux: Flux,  # flux instance
    c: Connection,  # connection instance
    cfn: str,  # full name of the connection instance
    icl: dict,  # list of initial conditions
    ind2: str,  # whitespace
    ind3: str,  # whitespace
) -> tuple(str, str):
    """Create equation string defining a flux that scales with the
    concentration in the upstream reservoir

    Example: M1_CG_D_b_to_L_b_TA_thc__F =
    M1.CG_D_b_to_L_b.TA_thc.scale * R[5]

    :param flux: Flux object
    :param c: connection instance
    :param cfn: full name of the connection instance
    :param icl: dict[Reservoir, list[int, int]] where reservoir
        indicates the reservoir handle, and the list contains the
        index into the reservoir data.  list[0] = concentration
        list[1] concentration of the light isotope.

    :returns: two strings with the respective equations for the change
              in the total reservoir concentration and the
              concentration of the light isotope
    """
    s_c = get_ic(c.source, icl)  # get index to concentration
    ex = f"{cfn}.scale * {s_c}"  # {c.id} scale with conc in {c.source.full_name}"
    exl = check_isotope_effects(ex, c, icl, ind3, ind2)
    ex, exl = check_signal_2(ex, exl, c)
    return ex, exl


def get_scale_with_flux_eq(
    flux: Flux,  # flux instance
    c: Connection,  # connection instance
    cfn: str,  # full name of the connection instance
    icl: dict,  # list of initial conditions
    ind2: str,  # indentation
    ind3: str,  # indentation
) -> tuple(str, str):
    """Equation defining a flux that scales with strength of another
    flux. If isotopes are used, use the isotope ratio of the upstream reservoir.

    :param flux: Flux object
    :param c: connection instance
    :param cfn: full name of the connection instance
    :param icl: dict[Reservoir, list[int, int]] where reservoir
        indicates the reservoir handle, and the list contains the
        index into the reservoir data.  list[0] = concentration
        list[1] concentration of the light isotope.

    :returns: two strings with the respective equations for the change
              in the total reservoir concentration and the
              concentration of the light isotope
    """

    # get the reference flux name
    p = flux.parent.ref_flux.parent
    fn = f"{p.full_name.replace('.', '_')}__F"
    # get the equation string for the flux
    ex = f"{cfn}.scale * {fn}"
    """ The flux for the light isotope will computed as follows:
    We will use the mass of the flux we or scaling, but that we will set the
    delta|alpha according to isotope ratio of the reference flux
    """
    exl = check_isotope_effects(ex, c, icl, ind3, ind2)
    ex, exl = check_signal_2(ex, exl, c)
    return ex, exl


def get_weathering_eq(
    flux: Flux,  # flux instance
    c: Connection,  # connection instance
    cfn: str,  # full name of the connection instance
    icl: dict,  # list of initial conditions
    ind2: str,
    ind3: str,
) -> tuple(str, str):
    """Equation defining a flux that scales with the concentration of pcO2

    F = F_0 (pcO2/pCO2_0) * c, see Zeebe, 2012, doi:10.5194/gmd-5-149-2012

    Example:

     M1_C_Fw_to_L_b_DIC_Ca_W__F = (
            M1.C_Fw_to_L_b_DIC_Ca_W.rate
            * M1.C_Fw_to_L_b_DIC_Ca_W.scale
            * (R[10]/M1.C_Fw_to_L_b_DIC_Ca_W.pco2_0)
            **  M1.C_Fw_to_L_b_DIC_Ca_W.ex
        )

    :param flux: Flux object
    :param c: connection instance
    :param cfn: full name of the connection instance
    :param icl: dict[Reservoir, list[int, int]] where reservoir
        indicates the reservoir handle, and the list contains the
        index into the reservoir data.  list[0] = concentration
        list[1] concentration of the light isotope.

    :returns: two strings with the respective equations for the change
              in the total reservoir concentration and the
              concentration of the light isotope

    """

    from esbmtk import Source, Sink

    s_c = get_ic(c.reservoir_ref, icl)
    ex = (
        f"(\n{ind3}{cfn}.rate\n"
        f"{ind3}* {cfn}.scale\n"
        f"{ind3}* ({s_c}/{cfn}.pco2_0)\n"
        f"{ind3}**  {cfn}.ex\n"
        f"{ind2})"
    )

    if c.isotopes:
        # get isotope ratio of source
        if isinstance(c.source, (Source, Sink)):
            s_c = f"{c.source.full_name}.c"
            s_l = f"{c.source.full_name}.l"
        else:
            s_c, s_l = get_ic(c.source, icl, True).replace(" ", "").split(",")

        fn = flux.full_name.replace(".", "_")
        exl = f"{fn} * {s_l} / {s_c}"
        exl = f"{exl}  # weathering + isotopes"
    else:
        exl = ""

    ex, exl = check_signal_2(ex, exl, c)

    return ex, exl


def get_gas_exchange_eq(
    flux: Flux,  # flux instance
    c: Connect,  # connection instance
    cfn: str,  # full name of the connection instance
    icl: dict,  # list of initial conditions
    ind2: str,
    ind3: str,
) -> str:
    """Equation defining a flux that scales with the concentration of
    of the gas in the atmosphere versus the concentration of the gas
    in solution.  See Zeebe, 2012, doi:10.5194/gmd-5-149-2012

    ..Example::

        M1_C_H_b_to_CO2_At_gex_hb_F = gas_exchange_ode(
             M1.C_H_b_to_CO2_At.scale,
             R[10],  # pco2 in atmosphere
             M1.C_H_b_to_CO2_At.water_vapor_pressure,
             M1.C_H_b_to_CO2_At.solubility,
             M1_H_b_CO2aq, # [co2]aq
         )  # gas_exchange

    :param flux: Flux object
    :param c: connection instance
    :param cfn: full name of the connection instance
    :param icl: dict[Reservoir, list[int, int]] where reservoir
        indicates the reservoir handle, and the list contains the
        index into the reservoir data.  list[0] = concentration
        list[1] concentration of the light isotope.

    :returns: string with the gas exchange equation
    """

    # get co2_aq reference
    lrn = f"{c.liquid_reservoir.full_name}"
    sp = lrn.split(".")[-1]
    # test if we refer to CO2 or other gas species
    if sp == "DIC":
        # here we reference the dyanamically calculated CO2aq from cs1,
        # and not the vector field of cs1
        refsp = f"{c.liquid_reservoir.parent.full_name}.CO2aq".replace(".", "_")
        # refsp = f"{c.liquid_reservoir.parent.full_name}.CO2aq"
    elif sp == "O2":
        s_c = get_ic(c.liquid_reservoir, icl)
        refsp = f"{s_c}"
    else:
        raise ValueError(f"Species{sp} has not definition for gex")

    # get atmosphere pcO2 reference
    pco2 = get_ic(c.gas_reservoir, icl)
    ex = (
        f"gas_exchange_ode(\n"
        f"{ind3}{cfn}.scale,\n"
        f"{ind3}{pco2},\n"
        f"{ind3}{cfn}.water_vapor_pressure,\n"
        f"{ind3}{cfn}.solubility,\n"
        f"{ind3}{refsp},\n"
        f"{ind2})"
    )
    ex, exl = check_signal_2(ex, "", c)
    ex = ex + "  # gas_exchange\n"

    return ex


def get_gas_exchange_w_isotopes_eq(
    flux: Flux,  # flux instance
    c: Connection,  # connection instance
    cfn: str,  # full name of the connection instance
    icl: dict,  # list of initial conditions
    ind2: str,
    ind3: str,
) -> tuple(str, str):
    """Equation defining a flux that scales with the concentration of
    of the gas in the atmosphere versus the concentration of the gas
    in solution.  See Zeebe, 2012, doi:10.5194/gmd-5-149-2012

    ..Example::

        M1_C_H_b_to_CO2_At_gex_hb_F, M1_C_H_b_to_CO2_At_gex_hb_F_l  = gas_exchange_ode(
           M1.C_H_b_to_CO2_At.scale, # surface area in m^2
           R[10],  # gas c in atmosphere
           R[11],  # gas c_l in atmosphere
           R[12],  # c of main species, e.g., DIC
           R[13],  # c_l of main species, e.g., DIC_12
           M1.C_H_b_to_CO2_At.water_vapor_pressure,
           M1.C_H_b_to_CO2_At.solubility,
           M1_H_b_CO2aq, # gas concentration in liquid, e.g., [co2]aq
           a_db,  # fractionation factor between dissolved CO2aq and HCO3-
           a_gb,  # fractionation between CO2g HCO3-
        )  # gas_exchange

    :param flux: Flux object
    :param c: connection instance
    :param cfn: full name of the connection instance
    :param icl: dict[Reservoir, list[int, int]] where reservoir
        indicates the reservoir handle, and the list contains the
        index into the reservoir data.  list[0] = concentration
        list[1] concentration of the light isotope.

    :returns: string with the gas exchange equation for the light
              isotope
    """

    # get isotope data
    pco2, pco2_l = get_ic(c.gas_reservoir, icl, True).replace(" ", "").split(",")
    dic, dic_l = get_ic(c.liquid_reservoir, icl, True).replace(" ", "").split(",")

    a_db = f"{c.liquid_reservoir.parent.full_name}.swc.a_db"
    a_dg = f"{c.liquid_reservoir.parent.full_name}.swc.a_dg"
    a_u = f"{c.liquid_reservoir.parent.full_name}.swc.a_u"

    # get co2_aq reference
    lrn = f"{c.liquid_reservoir.full_name}"
    sp = lrn.split(".")[-1]
    # test if we refer to CO2 or other gas species
    if sp == "DIC":
        # here we reference the dyanamically calculated CO2aq from cs1,
        # and not the vector field of cs1
        refsp = f"dCdt_{c.liquid_reservoir.parent.full_name}.CO2aq".replace(".", "_")
        # refsp = f"{c.liquid_reservoir.parent.full_name}.CO2aq"
    else:
        raise ValueError(f"Species{sp} has no isotope definition for gas exchange")

    ex = (
        f"gas_exchange_ode_with_isotopes(\n"
        f"{ind3}{cfn}.scale,\n"  # surface area in m^2
        f"{ind3}{pco2},\n"  # gas c in atmosphere
        f"{ind3}{pco2_l},\n"  # gas c_l in atmosphere
        f"{ind3}{dic},\n"  # c of reference species, e.g., DIC
        f"{ind3}{dic_l},\n"  # c_l reference species, e.g., DIC_12
        f"{ind3}{cfn}.water_vapor_pressure,\n"
        f"{ind3}{cfn}.solubility,\n"
        f"{ind3}{refsp},\n"  # gas concentration in liquid, e.g., [co2]aq
        f"{ind3}{a_db},\n"  # fractionation factor between dissolved CO2aq and HCO3-
        f"{ind3}{a_dg},\n"  # fractionation between CO2aq and CO2g
        f"{ind3}{a_u},\n"  # kinetic fractionation during gas exchange
        f"{ind2})"
    )
    ex = ex + "  # gas_exchange\n"

    return ex
